# Copyright (c) 2007 D-Level s.r.l. - All rights reserved

# Based on pynetsnmp-0.26.5 original code by Zenoss, Inc.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from threading import RLock


def is_instance_func(cls):
    cls.mutex.acquire()
    try:
        return hasattr(cls, "__singleton__")
    finally:
        cls.mutex.release()


class MetaSingleton(type):
    def __init__(self, *args, **kargs):
        self.mutex = RLock()
        return type.__init__(self, *args, **kargs)


class Singleton(object):
    """Singleton pattern implementation.
	
	Objects derived from this class will have a single instance,
	no matter how many time their constructors are called.
	"""
    __metaclass__ = MetaSingleton

    def __new__(cls, *args, **kargs):
        cls.mutex.acquire()
        try:
            s = cls.__dict__.get("__singleton__", None)
            if s is not None:
                return s
            cls.__singleton__ = s = object.__new__(cls)
            s.__destroyed = False
            s.init_class(*args, **kargs)
            return s
        finally:
            cls.mutex.release()

    def __init__(self, *args, **kargs):
        pass

    def init_class(self, *args, **kargs):
        """Class initialization method.
		
		Called only the first time the singleton's constructor
		is called.
		
		@param args: list of arguments.
		
		@param kargs: dictionary with arguments.
		"""
        pass

    def destroy(self):
        if '__singleton__' in self.__class__.__dict__:
            delattr(self.__class__, '__singleton__')
        self.__destroyed = True

    def __get_destroyed(self):
        return self.__destroyed

    destroyed = property(__get_destroyed)

    def __repr__(self):
        return '<%s@%d ()>' % (self.__class__.__name__, id(self))

    is_instance = classmethod(is_instance_func)
    # for compatibilty
    instantiated = is_instance



# Copyright (c) 2007 D-Level s.r.l. - All rights reserved
# THIS FILE WAS AUTO-GENERATED BY ConstantsGenerator.py - DO NOT EDIT!!!

#------- Default Constants -------

USM_LENGTH_OID_TRANSFORM = 10
NULL = None

#------- callback.h -------

MAX_CALLBACK_IDS = 2
MAX_CALLBACK_SUBIDS = 16
SNMP_CALLBACK_LIBRARY = 0
SNMP_CALLBACK_APPLICATION = 1
SNMP_CALLBACK_POST_READ_CONFIG = 0
SNMP_CALLBACK_STORE_DATA = 1
SNMP_CALLBACK_SHUTDOWN = 2
SNMP_CALLBACK_POST_PREMIB_READ_CONFIG = 3
SNMP_CALLBACK_LOGGING = 4
SNMP_CALLBACK_SESSION_INIT = 5
NETSNMP_CALLBACK_HIGHEST_PRIORITY = -1024
NETSNMP_CALLBACK_DEFAULT_PRIORITY = 0
NETSNMP_CALLBACK_LOWEST_PRIORITY = 1024

#------- asn1.h -------

PARSE_PACKET = 0
DUMP_PACKET = 1
MAX_SUBID = 0xFFFFFFFF
MAX_SUBID = 0xFF
MIN_OID_LEN = 2
MAX_OID_LEN = 128
MAX_NAME_LEN = MAX_OID_LEN
ASN_BOOLEAN = (0x01)
ASN_INTEGER = (0x02)
ASN_BIT_STR = (0x03)
ASN_OCTET_STR = (0x04)
ASN_NULL = (0x05)
ASN_OBJECT_ID = (0x06)
ASN_SEQUENCE = (0x10)
ASN_SET = (0x11)
ASN_UNIVERSAL = (0x00)
ASN_APPLICATION = (0x40)
ASN_CONTEXT = (0x80)
ASN_PRIVATE = (0xC0)
ASN_PRIMITIVE = (0x00)
ASN_CONSTRUCTOR = (0x20)
ASN_LONG_LEN = (0x80)
ASN_EXTENSION_ID = (0x1F)
ASN_BIT8 = (0x80)
ASN_OPAQUE_TAG1 = (ASN_CONTEXT | ASN_EXTENSION_ID)
ASN_OPAQUE_TAG2 = (0x30)
ASN_OPAQUE_TAG2U = (0x2f)
ASN_APP_OPAQUE = (ASN_APPLICATION | 4)
ASN_APP_COUNTER64 = (ASN_APPLICATION | 6)
ASN_APP_FLOAT = (ASN_APPLICATION | 8)
ASN_APP_DOUBLE = (ASN_APPLICATION | 9)
ASN_APP_I64 = (ASN_APPLICATION | 10)
ASN_APP_U64 = (ASN_APPLICATION | 11)
ASN_APP_UNION = (ASN_PRIVATE | 1)
ASN_OPAQUE_COUNTER64 = (ASN_OPAQUE_TAG2 + ASN_APP_COUNTER64)
ASN_OPAQUE_COUNTER64_MX_BER_LEN = 12
ASN_OPAQUE_FLOAT = (ASN_OPAQUE_TAG2 + ASN_APP_FLOAT)
ASN_OPAQUE_FLOAT_BER_LEN = 7
ASN_OPAQUE_DOUBLE = (ASN_OPAQUE_TAG2 + ASN_APP_DOUBLE)
ASN_OPAQUE_DOUBLE_BER_LEN = 11
ASN_OPAQUE_I64 = (ASN_OPAQUE_TAG2 + ASN_APP_I64)
ASN_OPAQUE_I64_MX_BER_LEN = 11
ASN_OPAQUE_U64 = (ASN_OPAQUE_TAG2 + ASN_APP_U64)
ASN_OPAQUE_U64_MX_BER_LEN = 12
ASN_PRIV_INCL_RANGE = (ASN_PRIVATE | 2)
ASN_PRIV_EXCL_RANGE = (ASN_PRIVATE | 3)
ASN_PRIV_DELEGATED = (ASN_PRIVATE | 5)
ASN_PRIV_IMPLIED_OCTET_STR = (ASN_PRIVATE | ASN_OCTET_STR)
ASN_PRIV_IMPLIED_OBJECT_ID = (ASN_PRIVATE | ASN_OBJECT_ID)
ASN_PRIV_RETRY = (ASN_PRIVATE | 7)

#------- snmp.h -------

SNMP_PORT = 161
SNMP_TRAP_PORT = 162
SNMP_MAX_LEN = 1500
SNMP_MIN_MAX_LEN = 484
SNMP_VERSION_1 = 0
SNMP_VERSION_2c = 1
SNMP_VERSION_2u = 2
SNMP_VERSION_3 = 3
SNMP_VERSION_sec = 128
SNMP_VERSION_2p = 129
SNMP_VERSION_2star = 130
SNMP_MSG_GET = (ASN_CONTEXT | ASN_CONSTRUCTOR | 0x0)
SNMP_MSG_GETNEXT = (ASN_CONTEXT | ASN_CONSTRUCTOR | 0x1)
SNMP_MSG_RESPONSE = (ASN_CONTEXT | ASN_CONSTRUCTOR | 0x2)
SNMP_MSG_SET = (ASN_CONTEXT | ASN_CONSTRUCTOR | 0x3)
SNMP_MSG_TRAP = (ASN_CONTEXT | ASN_CONSTRUCTOR | 0x4)
SNMP_MSG_GETBULK = (ASN_CONTEXT | ASN_CONSTRUCTOR | 0x5)
SNMP_MSG_INFORM = (ASN_CONTEXT | ASN_CONSTRUCTOR | 0x6)
SNMP_MSG_TRAP2 = (ASN_CONTEXT | ASN_CONSTRUCTOR | 0x7)
SNMP_MSG_REPORT = (ASN_CONTEXT | ASN_CONSTRUCTOR | 0x8)
SNMP_MSG_INTERNAL_SET_BEGIN = -1
SNMP_MSG_INTERNAL_SET_RESERVE1 = 0
SNMP_MSG_INTERNAL_SET_RESERVE2 = 1
SNMP_MSG_INTERNAL_SET_ACTION = 2
SNMP_MSG_INTERNAL_SET_COMMIT = 3
SNMP_MSG_INTERNAL_SET_FREE = 4
SNMP_MSG_INTERNAL_SET_UNDO = 5
SNMP_MSG_INTERNAL_SET_MAX = 6
SNMP_MSG_INTERNAL_CHECK_VALUE = 17
SNMP_MSG_INTERNAL_ROW_CREATE = 18
SNMP_MSG_INTERNAL_UNDO_SETUP = 19
SNMP_MSG_INTERNAL_SET_VALUE = 20
SNMP_MSG_INTERNAL_CHECK_CONSISTENCY = 21
SNMP_MSG_INTERNAL_UNDO_SET = 22
SNMP_MSG_INTERNAL_COMMIT = 23
SNMP_MSG_INTERNAL_UNDO_COMMIT = 24
SNMP_MSG_INTERNAL_IRREVERSIBLE_COMMIT = 25
SNMP_MSG_INTERNAL_UNDO_CLEANUP = 26
SNMP_MSG_INTERNAL_PRE_REQUEST = 128
SNMP_MSG_INTERNAL_OBJECT_LOOKUP = 129
SNMP_MSG_INTERNAL_POST_REQUEST = 130
SNMP_MSG_INTERNAL_GET_STASH = 131
SNMP_NOSUCHOBJECT = (ASN_CONTEXT | ASN_PRIMITIVE | 0x0)
SNMP_NOSUCHINSTANCE = (ASN_CONTEXT | ASN_PRIMITIVE | 0x1)
SNMP_ENDOFMIBVIEW = (ASN_CONTEXT | ASN_PRIMITIVE | 0x2)
SNMP_ERR_NOERROR = (0)
SNMP_ERR_TOOBIG = (1)
SNMP_ERR_NOSUCHNAME = (2)
SNMP_ERR_BADVALUE = (3)
SNMP_ERR_READONLY = (4)
SNMP_ERR_GENERR = (5)
SNMP_ERR_NOACCESS = (6)
SNMP_ERR_WRONGTYPE = (7)
SNMP_ERR_WRONGLENGTH = (8)
SNMP_ERR_WRONGENCODING = (9)
SNMP_ERR_WRONGVALUE = (10)
SNMP_ERR_NOCREATION = (11)
SNMP_ERR_INCONSISTENTVALUE = (12)
SNMP_ERR_RESOURCEUNAVAILABLE = (13)
SNMP_ERR_COMMITFAILED = (14)
SNMP_ERR_UNDOFAILED = (15)
SNMP_ERR_AUTHORIZATIONERROR = (16)
SNMP_ERR_NOTWRITABLE = (17)
SNMP_ERR_INCONSISTENTNAME = (18)
MAX_SNMP_ERR = 18
SNMP_TRAP_COLDSTART = (0)
SNMP_TRAP_WARMSTART = (1)
SNMP_TRAP_LINKDOWN = (2)
SNMP_TRAP_LINKUP = (3)
SNMP_TRAP_AUTHFAIL = (4)
SNMP_TRAP_EGPNEIGHBORLOSS = (5)
SNMP_TRAP_ENTERPRISESPECIFIC = (6)
SNMP_ROW_NONEXISTENT = 0
SNMP_ROW_ACTIVE = 1
SNMP_ROW_NOTINSERVICE = 2
SNMP_ROW_NOTREADY = 3
SNMP_ROW_CREATEANDGO = 4
SNMP_ROW_CREATEANDWAIT = 5
SNMP_ROW_DESTROY = 6
SNMP_STORAGE_NONE = 0
SNMP_STORAGE_OTHER = 1
SNMP_STORAGE_VOLATILE = 2
SNMP_STORAGE_NONVOLATILE = 3
SNMP_STORAGE_PERMANENT = 4
SNMP_STORAGE_READONLY = 5
SNMP_MP_MODEL_SNMPv1 = 0
SNMP_MP_MODEL_SNMPv2c = 1
SNMP_MP_MODEL_SNMPv2u = 2
SNMP_MP_MODEL_SNMPv3 = 3
SNMP_MP_MODEL_SNMPv2p = 256
SNMP_SEC_MODEL_ANY = 0
SNMP_SEC_MODEL_SNMPv1 = 1
SNMP_SEC_MODEL_SNMPv2c = 2
SNMP_SEC_MODEL_USM = 3
SNMP_SEC_MODEL_SNMPv2p = 256
SNMP_SEC_LEVEL_NOAUTH = 1
SNMP_SEC_LEVEL_AUTHNOPRIV = 2
SNMP_SEC_LEVEL_AUTHPRIV = 3
SNMP_MSG_FLAG_AUTH_BIT = 0x01
SNMP_MSG_FLAG_PRIV_BIT = 0x02
SNMP_MSG_FLAG_RPRT_BIT = 0x04
UCD_MSG_FLAG_RESPONSE_PDU = 0x100
UCD_MSG_FLAG_EXPECT_RESPONSE = 0x200
UCD_MSG_FLAG_FORCE_PDU_COPY = 0x400
UCD_MSG_FLAG_ALWAYS_IN_VIEW = 0x800
UCD_MSG_FLAG_PDU_TIMEOUT = 0x1000
UCD_MSG_FLAG_ONE_PASS_ONLY = 0x2000
UCD_MSG_FLAG_TUNNELED = 0x4000
SNMP_VIEW_INCLUDED = 1
SNMP_VIEW_EXCLUDED = 2
SNMP_OID_INTERNET = 1, 3, 6, 1
SNMP_OID_ENTERPRISES = SNMP_OID_INTERNET, 4, 1
SNMP_OID_MIB2 = SNMP_OID_INTERNET, 2, 1
SNMP_OID_SNMPV2 = SNMP_OID_INTERNET, 6
SNMP_OID_SNMPMODULES = SNMP_OID_SNMPV2, 3
SNMPADMINLENGTH = 255

#------- snmp_api.h -------

USM_AUTH_KU_LEN = 32
USM_PRIV_KU_LEN = 32
SNMP_DEFAULT_COMMUNITY_LEN = 0
SNMP_DEFAULT_RETRIES = -1
SNMP_DEFAULT_TIMEOUT = -1
SNMP_DEFAULT_REMPORT = 0
SNMP_DEFAULT_REQID = -1
SNMP_DEFAULT_MSGID = -1
SNMP_DEFAULT_ERRSTAT = -1
SNMP_DEFAULT_ERRINDEX = -1
SNMP_DEFAULT_ADDRESS = 0
SNMP_DEFAULT_PEERNAME = NULL
SNMP_DEFAULT_ENTERPRISE_LENGTH = 0
SNMP_DEFAULT_TIME = 0
SNMP_DEFAULT_VERSION = -1
SNMP_DEFAULT_SECMODEL = -1
SNMP_DEFAULT_CONTEXT = ""
SNMP_DEFAULT_AUTH_PROTOLEN = USM_LENGTH_OID_TRANSFORM
SNMP_DEFAULT_PRIV_PROTOLEN = USM_LENGTH_OID_TRANSFORM
SNMP_MAX_MSG_SIZE = 1472
SNMP_MAX_MSG_V3_HDRS = (4 + 3 + 4 + 7 + 7 + 3 + 7 + 16)
SNMP_MAX_ENG_SIZE = 32
SNMP_MAX_SEC_NAME_SIZE = 256
SNMP_MAX_CONTEXT_SIZE = 256
SNMP_SEC_PARAM_BUF_SIZE = 256
SNMPV3_IGNORE_UNAUTH_REPORTS = 0
SNMP_SESS_NONAUTHORITATIVE = 0
SNMP_SESS_AUTHORITATIVE = 1
SNMP_SESS_UNKNOWNAUTH = 2
REPORT_STATS_LEN = 9
REPORT_snmpUnknownSecurityModels_NUM = 1
REPORT_snmpInvalidMsgs_NUM = 2
REPORT_usmStatsUnsupportedSecLevels_NUM = 1
REPORT_usmStatsNotInTimeWindows_NUM = 2
REPORT_usmStatsUnknownUserNames_NUM = 3
REPORT_usmStatsUnknownEngineIDs_NUM = 4
REPORT_usmStatsWrongDigests_NUM = 5
REPORT_usmStatsDecryptionErrors_NUM = 6
SNMP_DETAIL_SIZE = 512
SNMP_FLAGS_USER_CREATED = 0x200
SNMP_FLAGS_DONT_PROBE = 0x100
SNMP_FLAGS_STREAM_SOCKET = 0x80
SNMP_FLAGS_LISTENING = 0x40
SNMP_FLAGS_SUBSESSION = 0x20
SNMP_FLAGS_STRIKE2 = 0x02
SNMP_FLAGS_STRIKE1 = 0x01
SNMPERR_SUCCESS = (0)
SNMPERR_GENERR = (-1)
SNMPERR_BAD_LOCPORT = (-2)
SNMPERR_BAD_ADDRESS = (-3)
SNMPERR_BAD_SESSION = (-4)
SNMPERR_TOO_LONG = (-5)
SNMPERR_NO_SOCKET = (-6)
SNMPERR_V2_IN_V1 = (-7)
SNMPERR_V1_IN_V2 = (-8)
SNMPERR_BAD_REPEATERS = (-9)
SNMPERR_BAD_REPETITIONS = (-10)
SNMPERR_BAD_ASN1_BUILD = (-11)
SNMPERR_BAD_SENDTO = (-12)
SNMPERR_BAD_PARSE = (-13)
SNMPERR_BAD_VERSION = (-14)
SNMPERR_BAD_SRC_PARTY = (-15)
SNMPERR_BAD_DST_PARTY = (-16)
SNMPERR_BAD_CONTEXT = (-17)
SNMPERR_BAD_COMMUNITY = (-18)
SNMPERR_NOAUTH_DESPRIV = (-19)
SNMPERR_BAD_ACL = (-20)
SNMPERR_BAD_PARTY = (-21)
SNMPERR_ABORT = (-22)
SNMPERR_UNKNOWN_PDU = (-23)
SNMPERR_TIMEOUT = (-24)
SNMPERR_BAD_RECVFROM = (-25)
SNMPERR_BAD_ENG_ID = (-26)
SNMPERR_BAD_SEC_NAME = (-27)
SNMPERR_BAD_SEC_LEVEL = (-28)
SNMPERR_ASN_PARSE_ERR = (-29)
SNMPERR_UNKNOWN_SEC_MODEL = (-30)
SNMPERR_INVALID_MSG = (-31)
SNMPERR_UNKNOWN_ENG_ID = (-32)
SNMPERR_UNKNOWN_USER_NAME = (-33)
SNMPERR_UNSUPPORTED_SEC_LEVEL = (-34)
SNMPERR_AUTHENTICATION_FAILURE = (-35)
SNMPERR_NOT_IN_TIME_WINDOW = (-36)
SNMPERR_DECRYPTION_ERR = (-37)
SNMPERR_SC_GENERAL_FAILURE = (-38)
SNMPERR_SC_NOT_CONFIGURED = (-39)
SNMPERR_KT_NOT_AVAILABLE = (-40)
SNMPERR_UNKNOWN_REPORT = (-41)
SNMPERR_USM_GENERICERROR = (-42)
SNMPERR_USM_UNKNOWNSECURITYNAME = (-43)
SNMPERR_USM_UNSUPPORTEDSECURITYLEVEL = (-44)
SNMPERR_USM_ENCRYPTIONERROR = (-45)
SNMPERR_USM_AUTHENTICATIONFAILURE = (-46)
SNMPERR_USM_PARSEERROR = (-47)
SNMPERR_USM_UNKNOWNENGINEID = (-48)
SNMPERR_USM_NOTINTIMEWINDOW = (-49)
SNMPERR_USM_DECRYPTIONERROR = (-50)
SNMPERR_NOMIB = (-51)
SNMPERR_RANGE = (-52)
SNMPERR_MAX_SUBID = (-53)
SNMPERR_BAD_SUBID = (-54)
SNMPERR_LONG_OID = (-55)
SNMPERR_BAD_NAME = (-56)
SNMPERR_VALUE = (-57)
SNMPERR_UNKNOWN_OBJID = (-58)
SNMPERR_NULL_PDU = (-59)
SNMPERR_NO_VARS = (-60)
SNMPERR_VAR_TYPE = (-61)
SNMPERR_MALLOC = (-62)
SNMPERR_KRB5 = (-63)
SNMPERR_PROTOCOL = (-64)
SNMPERR_OID_NONINCREASING = (-65)
SNMPERR_MAX = (-65)
NETSNMP_CALLBACK_OP_RECEIVED_MESSAGE = 1
NETSNMP_CALLBACK_OP_TIMED_OUT = 2
NETSNMP_CALLBACK_OP_SEND_FAILED = 3
NETSNMP_CALLBACK_OP_CONNECT = 4
NETSNMP_CALLBACK_OP_DISCONNECT = 5
STAT_SNMPUNKNOWNSECURITYMODELS = 0
STAT_SNMPINVALIDMSGS = 1
STAT_SNMPUNKNOWNPDUHANDLERS = 2
STAT_MPD_STATS_START = STAT_SNMPUNKNOWNSECURITYMODELS
STAT_MPD_STATS_END = STAT_SNMPUNKNOWNPDUHANDLERS
STAT_USMSTATSUNSUPPORTEDSECLEVELS = 3
STAT_USMSTATSNOTINTIMEWINDOWS = 4
STAT_USMSTATSUNKNOWNUSERNAMES = 5
STAT_USMSTATSUNKNOWNENGINEIDS = 6
STAT_USMSTATSWRONGDIGESTS = 7
STAT_USMSTATSDECRYPTIONERRORS = 8
STAT_USM_STATS_START = STAT_USMSTATSUNSUPPORTEDSECLEVELS
STAT_USM_STATS_END = STAT_USMSTATSDECRYPTIONERRORS
STAT_SNMPINPKTS = 9
STAT_SNMPOUTPKTS = 10
STAT_SNMPINBADVERSIONS = 11
STAT_SNMPINBADCOMMUNITYNAMES = 12
STAT_SNMPINBADCOMMUNITYUSES = 13
STAT_SNMPINASNPARSEERRS = 14
STAT_SNMPINTOOBIGS = 16
STAT_SNMPINNOSUCHNAMES = 17
STAT_SNMPINBADVALUES = 18
STAT_SNMPINREADONLYS = 19
STAT_SNMPINGENERRS = 20
STAT_SNMPINTOTALREQVARS = 21
STAT_SNMPINTOTALSETVARS = 22
STAT_SNMPINGETREQUESTS = 23
STAT_SNMPINGETNEXTS = 24
STAT_SNMPINSETREQUESTS = 25
STAT_SNMPINGETRESPONSES = 26
STAT_SNMPINTRAPS = 27
STAT_SNMPOUTTOOBIGS = 28
STAT_SNMPOUTNOSUCHNAMES = 29
STAT_SNMPOUTBADVALUES = 30
STAT_SNMPOUTGENERRS = 32
STAT_SNMPOUTGETREQUESTS = 33
STAT_SNMPOUTGETNEXTS = 34
STAT_SNMPOUTSETREQUESTS = 35
STAT_SNMPOUTGETRESPONSES = 36
STAT_SNMPOUTTRAPS = 37
STAT_SNMPSILENTDROPS = 39
STAT_SNMPPROXYDROPS = 40
STAT_SNMP_STATS_START = STAT_SNMPINPKTS
STAT_SNMP_STATS_END = STAT_SNMPPROXYDROPS
STAT_SNMPUNAVAILABLECONTEXTS = 41
STAT_SNMPUNKNOWNCONTEXTS = 42
STAT_TARGET_STATS_START = STAT_SNMPUNAVAILABLECONTEXTS
STAT_TARGET_STATS_END = STAT_SNMPUNKNOWNCONTEXTS
MAX_STATS = 43

#------- snmp_impl.h -------

COMMUNITY_MAX_LEN = 256
SPRINT_MAX_LEN = 2560
NULL = 0
TRUE = 1
FALSE = 0
READ = 1
WRITE = 0
RESERVE1 = 0
RESERVE2 = 1
ACTION = 2
COMMIT = 3
FREE = 4
UNDO = 5
FINISHED_SUCCESS = 9
FINISHED_FAILURE = 10
RONLY = 0x1
RWRITE = 0x2
NOACCESS = 0x0000
ASN_IPADDRESS = (ASN_APPLICATION | 0)
ASN_COUNTER = (ASN_APPLICATION | 1)
ASN_GAUGE = (ASN_APPLICATION | 2)
ASN_UNSIGNED = (ASN_APPLICATION | 2)
ASN_TIMETICKS = (ASN_APPLICATION | 3)
ASN_OPAQUE = (ASN_APPLICATION | 4)
ASN_NSAP = (ASN_APPLICATION | 5)
ASN_COUNTER64 = (ASN_APPLICATION | 6)
ASN_UINTEGER = (ASN_APPLICATION | 7)
ASN_FLOAT = (ASN_APPLICATION | 8)
ASN_DOUBLE = (ASN_APPLICATION | 9)
ASN_INTEGER64 = (ASN_APPLICATION | 10)
ASN_UNSIGNED64 = (ASN_APPLICATION | 11)
FIRST_PASS = 1
LAST_PASS = 2

#------- snmp_logging.h -------

LOG_EMERG = 0
LOG_ALERT = 1
LOG_CRIT = 2
LOG_ERR = 3
LOG_WARNING = 4
LOG_NOTICE = 5
LOG_INFO = 6
LOG_DEBUG = 7
DEFAULT_LOG_ID = "net-snmp"
NETSNMP_LOGHANDLER_STDOUT = 1
NETSNMP_LOGHANDLER_STDERR = 2
NETSNMP_LOGHANDLER_FILE = 3
NETSNMP_LOGHANDLER_SYSLOG = 4
NETSNMP_LOGHANDLER_CALLBACK = 5
NETSNMP_LOGHANDLER_NONE = 6

#------- snmp_client.h -------

STAT_SUCCESS = 0
STAT_ERROR = 1
STAT_TIMEOUT = 2

#------- mib.h -------

MIB = 1, 3, 6, 1, 2, 1
MIB_IFTYPE_OTHER = 1
MIB_IFTYPE_REGULAR1822 = 2
MIB_IFTYPE_HDH1822 = 3
MIB_IFTYPE_DDNX25 = 4
MIB_IFTYPE_RFC877X25 = 5
MIB_IFTYPE_ETHERNETCSMACD = 6
MIB_IFTYPE_ISO88023CSMACD = 7
MIB_IFTYPE_ISO88024TOKENBUS = 8
MIB_IFTYPE_ISO88025TOKENRING = 9
MIB_IFTYPE_ISO88026MAN = 10
MIB_IFTYPE_STARLAN = 11
MIB_IFTYPE_PROTEON10MBIT = 12
MIB_IFTYPE_PROTEON80MBIT = 13
MIB_IFTYPE_HYPERCHANNEL = 14
MIB_IFTYPE_FDDI = 15
MIB_IFTYPE_LAPB = 16
MIB_IFTYPE_SDLC = 17
MIB_IFTYPE_T1CARRIER = 18
MIB_IFTYPE_CEPT = 19
MIB_IFTYPE_BASICISDN = 20
MIB_IFTYPE_PRIMARYISDN = 21
MIB_IFTYPE_PROPPOINTTOPOINTSERIAL = 22
MIB_IFSTATUS_UP = 1
MIB_IFSTATUS_DOWN = 2
MIB_IFSTATUS_TESTING = 3
MIB_FORWARD_GATEWAY = 1
MIB_FORWARD_HOST = 2
MIB_IPROUTETYPE_OTHER = 1
MIB_IPROUTETYPE_INVALID = 2
MIB_IPROUTETYPE_DIRECT = 3
MIB_IPROUTETYPE_REMOTE = 4
MIB_IPROUTEPROTO_OTHER = 1
MIB_IPROUTEPROTO_LOCAL = 2
MIB_IPROUTEPROTO_NETMGMT = 3
MIB_IPROUTEPROTO_ICMP = 4
MIB_IPROUTEPROTO_EGP = 5
MIB_IPROUTEPROTO_GGP = 6
MIB_IPROUTEPROTO_HELLO = 7
MIB_IPROUTEPROTO_RIP = 8
MIB_IPROUTEPROTO_ISIS = 9
MIB_IPROUTEPROTO_ESIS = 10
MIB_IPROUTEPROTO_CISCOIGRP = 11
MIB_IPROUTEPROTO_BBNSPFIGP = 12
MIB_IPROUTEPROTO_OIGP = 13
MIB_TCPRTOALG_OTHER = 1
MIB_TCPRTOALG_CONSTANT = 2
MIB_TCPRTOALG_RSRE = 3
MIB_TCPRTOALG_VANJ = 4
MIB_TCPCONNSTATE_CLOSED = 1
MIB_TCPCONNSTATE_LISTEN = 2
MIB_TCPCONNSTATE_SYNSENT = 3
MIB_TCPCONNSTATE_SYNRECEIVED = 4
MIB_TCPCONNSTATE_ESTABLISHED = 5
MIB_TCPCONNSTATE_FINWAIT1 = 6
MIB_TCPCONNSTATE_FINWAIT2 = 7
MIB_TCPCONNSTATE_CLOSEWAIT = 8
MIB_TCPCONNSTATE_LASTACK = 9
MIB_TCPCONNSTATE_CLOSING = 10
MIB_TCPCONNSTATE_TIMEWAIT = 11
MIB_EGPNEIGHSTATE_IDLE = 1
MIB_EGPNEIGHSTATE_AQUISITION = 2
MIB_EGPNEIGHSTATE_DOWN = 3
MIB_EGPNEIGHSTATE_UP = 4
MIB_EGPNEIGHSTATE_CEASE = 5
NETSNMP_STRING_OUTPUT_GUESS = 1
NETSNMP_STRING_OUTPUT_ASCII = 2
NETSNMP_STRING_OUTPUT_HEX = 3
NETSNMP_OID_OUTPUT_SUFFIX = 1
NETSNMP_OID_OUTPUT_MODULE = 2
NETSNMP_OID_OUTPUT_FULL = 3
NETSNMP_OID_OUTPUT_NUMERIC = 4
NETSNMP_OID_OUTPUT_UCD = 5
NETSNMP_OID_OUTPUT_NONE = 6

#------- default_store.h -------

NETSNMP_DS_MAX_IDS = 3
NETSNMP_DS_MAX_SUBIDS = 40
NETSNMP_DS_LIBRARY_ID = 0
NETSNMP_DS_APPLICATION_ID = 1
NETSNMP_DS_TOKEN_ID = 2
NETSNMP_DS_LIB_MIB_ERRORS = 0
NETSNMP_DS_LIB_SAVE_MIB_DESCRS = 1
NETSNMP_DS_LIB_MIB_COMMENT_TERM = 2
NETSNMP_DS_LIB_MIB_PARSE_LABEL = 3
NETSNMP_DS_LIB_DUMP_PACKET = 4
NETSNMP_DS_LIB_LOG_TIMESTAMP = 5
NETSNMP_DS_LIB_DONT_READ_CONFIGS = 6
NETSNMP_DS_LIB_DISABLE_CONFIG_LOAD = NETSNMP_DS_LIB_DONT_READ_CONFIGS
NETSNMP_DS_LIB_MIB_REPLACE = 7
NETSNMP_DS_LIB_PRINT_NUMERIC_ENUM = 8
NETSNMP_DS_LIB_PRINT_NUMERIC_OIDS = 9
NETSNMP_DS_LIB_DONT_BREAKDOWN_OIDS = 10
NETSNMP_DS_LIB_ALARM_DONT_USE_SIG = 11
NETSNMP_DS_LIB_PRINT_FULL_OID = 12
NETSNMP_DS_LIB_QUICK_PRINT = 13
NETSNMP_DS_LIB_RANDOM_ACCESS = 14
NETSNMP_DS_LIB_REGEX_ACCESS = 15
NETSNMP_DS_LIB_DONT_CHECK_RANGE = 16
NETSNMP_DS_LIB_NO_TOKEN_WARNINGS = 17
NETSNMP_DS_LIB_NUMERIC_TIMETICKS = 18
NETSNMP_DS_LIB_ESCAPE_QUOTES = 19
NETSNMP_DS_LIB_REVERSE_ENCODE = 20
NETSNMP_DS_LIB_PRINT_BARE_VALUE = 21
NETSNMP_DS_LIB_EXTENDED_INDEX = 22
NETSNMP_DS_LIB_PRINT_HEX_TEXT = 23
NETSNMP_DS_LIB_PRINT_UCD_STYLE_OID = 24
NETSNMP_DS_LIB_READ_UCD_STYLE_OID = 25
NETSNMP_DS_LIB_HAVE_READ_PREMIB_CONFIG = 26
NETSNMP_DS_LIB_HAVE_READ_CONFIG = 27
NETSNMP_DS_LIB_QUICKE_PRINT = 28
NETSNMP_DS_LIB_DONT_PRINT_UNITS = 29
NETSNMP_DS_LIB_NO_DISPLAY_HINT = 30
NETSNMP_DS_LIB_16BIT_IDS = 31
NETSNMP_DS_LIB_DONT_PERSIST_STATE = 32
NETSNMP_DS_LIB_2DIGIT_HEX_OUTPUT = 33
NETSNMP_DS_LIB_IGNORE_NO_COMMUNITY = 34
NETSNMP_DS_LIB_DISABLE_PERSISTENT_LOAD = 35
NETSNMP_DS_LIB_DISABLE_PERSISTENT_SAVE = 36
NETSNMP_DS_LIB_APPEND_LOGFILES = 37
NETSNMP_DS_LIB_MIB_WARNINGS = 0
NETSNMP_DS_LIB_SECLEVEL = 1
NETSNMP_DS_LIB_SNMPVERSION = 2
NETSNMP_DS_LIB_DEFAULT_PORT = 3
NETSNMP_DS_LIB_OID_OUTPUT_FORMAT = 4
NETSNMP_DS_LIB_PRINT_SUFFIX_ONLY = NETSNMP_DS_LIB_OID_OUTPUT_FORMAT
NETSNMP_DS_LIB_STRING_OUTPUT_FORMAT = 5
NETSNMP_DS_LIB_HEX_OUTPUT_LENGTH = 6
NETSNMP_DS_LIB_SERVERSENDBUF = 7
NETSNMP_DS_LIB_SERVERRECVBUF = 8
NETSNMP_DS_LIB_CLIENTSENDBUF = 9
NETSNMP_DS_LIB_CLIENTRECVBUF = 10
NETSNMP_DS_SNMP_VERSION_1 = 128
NETSNMP_DS_SNMP_VERSION_2c = 1
NETSNMP_DS_SNMP_VERSION_3 = 3
NETSNMP_DS_LIB_SECNAME = 0
NETSNMP_DS_LIB_CONTEXT = 1
NETSNMP_DS_LIB_PASSPHRASE = 2
NETSNMP_DS_LIB_AUTHPASSPHRASE = 3
NETSNMP_DS_LIB_PRIVPASSPHRASE = 4
NETSNMP_DS_LIB_OPTIONALCONFIG = 5
NETSNMP_DS_LIB_APPTYPE = 6
NETSNMP_DS_LIB_COMMUNITY = 7
NETSNMP_DS_LIB_PERSISTENT_DIR = 8
NETSNMP_DS_LIB_CONFIGURATION_DIR = 9
NETSNMP_DS_LIB_SECMODEL = 10
NETSNMP_DS_LIB_MIBDIRS = 11
NETSNMP_DS_LIB_OIDSUFFIX = 12
NETSNMP_DS_LIB_OIDPREFIX = 13
NETSNMP_DS_LIB_CLIENT_ADDR = 14
NETSNMP_DS_LIB_TEMP_FILE_PATTERN = 15
NETSNMP_DS_LIB_AUTHMASTERKEY = 16
NETSNMP_DS_LIB_PRIVMASTERKEY = 17
NETSNMP_DS_LIB_AUTHLOCALIZEDKEY = 18
NETSNMP_DS_LIB_PRIVLOCALIZEDKEY = 19
NETSNMP_DS_LIB_APPTYPES = 20
NETSNMP_DS_LIB_KSM_KEYTAB = 21
NETSNMP_DS_LIB_KSM_SERVICE_NAME = 22

#------- parse.h -------

MAXLABEL = 64
MAXTOKEN = 128
MAXQUOTESTR = 4096
TYPE_OTHER = 0
TYPE_OBJID = 1
TYPE_OCTETSTR = 2
TYPE_INTEGER = 3
TYPE_NETADDR = 4
TYPE_IPADDR = 5
TYPE_COUNTER = 6
TYPE_GAUGE = 7
TYPE_TIMETICKS = 8
TYPE_OPAQUE = 9
TYPE_NULL = 10
TYPE_COUNTER64 = 11
TYPE_BITSTRING = 12
TYPE_NSAPADDRESS = 13
TYPE_UINTEGER = 14
TYPE_UNSIGNED32 = 15
TYPE_INTEGER32 = 16
TYPE_SIMPLE_LAST = 16
TYPE_TRAPTYPE = 20
TYPE_NOTIFTYPE = 21
TYPE_OBJGROUP = 22
TYPE_NOTIFGROUP = 23
TYPE_MODID = 24
TYPE_AGENTCAP = 25
TYPE_MODCOMP = 26
TYPE_OBJIDENTITY = 27
MIB_ACCESS_READONLY = 18
MIB_ACCESS_READWRITE = 19
MIB_ACCESS_WRITEONLY = 20
MIB_ACCESS_NOACCESS = 21
MIB_ACCESS_NOTIFY = 67
MIB_ACCESS_CREATE = 48
MIB_STATUS_MANDATORY = 23
MIB_STATUS_OPTIONAL = 24
MIB_STATUS_OBSOLETE = 25
MIB_STATUS_DEPRECATED = 39
MIB_STATUS_CURRENT = 57
ANON = "anonymous#"



# Copyright (c) 2007 D-Level s.r.l. - All rights reserved

# Based on pynetsnmp-0.26.5 original code by Zenoss, Inc.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

__all__ = [
    'SNMPManager',
    'SNMPSession',
    'SnmpError',
    'str_to_oid',
    'strs_to_oids',
    'oid_to_str',
    'oids_to_strs',
    'oid_to_dot',
    'oids_to_dots',
    'get_oid_info',
    'get_snmp_error',
    'mkoid',
    'create_string_buffer',
    'lib',
    'get_enum_dict',
    'setup_library',
    'add_mib_dir',
    'read_module',
    'init_mib',
]

# pylint: disable=unused-wildcard-import,invalid-name
#----------------------------------------------------------------------------

import os
import sys
import time
from select import select
import threading
import traceback

import struct
from ctypes import *
from ctypes.util import find_library

from .constants import *

#----------------------------------------------------------------------------

from .Singleton import Singleton

#----------------------------------------------------------------------------


def syncronized(method, mutex_finder=lambda x: x.mutex):
    def m(self, *args, **kargs):
        mutex = mutex_finder(self)

        mutex.acquire()
        try:
            return method(self, *args, **kargs)
        finally:
            mutex.release()

    m.real_method = method
    return m


#----------------------------------------------------------------------------
# load libraries

# freebsd stuff
if sys.platform.find('free') > -1:
    find_library_orig = find_library

    def find_library(name):
        for name in [
                '/usr/lib/lib%s.so' % name, '/usr/local/lib/lib%s.so' % name
        ]:
            if os.path.exists(name):
                return name
        return find_library_orig(name)


try:
    crypto = CDLL(find_library('crypto'), RTLD_GLOBAL)
except Exception:
    pass

lib_path = find_library('netsnmp')

lib = CDLL(lib_path)
lib.netsnmp_get_version.restype = c_char_p

#----------------------------------------------------------------------------
# type definitions

c_int_p = c_void_p
authenticator = CFUNCTYPE(c_char_p, c_int_p, c_char_p, c_int)
oid = c_long
u_long = c_ulong
u_short = c_ushort
u_char_p = c_char_p
u_int = c_uint
size_t = c_size_t
u_char = c_byte


class netsnmp_session(Structure):
    pass


class netsnmp_pdu(Structure):
    pass


# int (*netsnmp_callback) (int, netsnmp_session *, int, netsnmp_pdu *, void *);
netsnmp_callback = CFUNCTYPE(c_int, c_int,
                             POINTER(netsnmp_session), c_int,
                             POINTER(netsnmp_pdu), c_void_p)

version = lib.netsnmp_get_version()
float_version = float('.'.join(version.split('.')[:2]))
localname = []
param_name = []

if float_version < 5.7:
    raise ImportError("netsnmp version 5.7 or greater is required")

class netsnmp_container_s(Structure): pass

netsnmp_session._fields_ = [
    ('version', c_long),
    ('retries', c_int),
    ('timeout', c_long),
    ('flags', u_long),
    ('subsession', POINTER(netsnmp_session)),
    ('next', POINTER(netsnmp_session)),
    ('peername', c_char_p),
    ('remote_port', u_short),
    ('localname', c_char_p),
    ('local_port', u_short),
    ('authenticator', authenticator),
    ('callback', netsnmp_callback),
    ('callback_magic', c_void_p),
    ('s_errno', c_int),
    ('s_snmp_errno', c_int),
    ('sessid', c_long),
    ('community', u_char_p),
    ('community_len', size_t),
    ('rcvMsgMaxSize', size_t),
    ('sndMsgMaxSize', size_t),
    ('isAuthoritative', u_char),
    ('contextEngineID', u_char_p),
    ('contextEngineIDLen', size_t),
    ('engineBoots', u_int),
    ('engineTime', u_int),
    ('contextName', c_char_p),
    ('contextNameLen', size_t),
    ('securityEngineID', u_char_p),
    ('securityEngineIDLen', size_t),
    ('securityName', c_char_p),
    ('securityNameLen', size_t),
    ('securityAuthProto', POINTER(oid)),
    ('securityAuthProtoLen', size_t),
    ('securityAuthKey', u_char * USM_AUTH_KU_LEN),
    ('securityAuthKeyLen', c_size_t),
    ('securityAuthLocalKey', c_char_p),
    ('securityAuthLocalKeyLen', c_size_t),
    ('securityPrivProto', POINTER(oid)),
    ('securityPrivProtoLen', c_size_t),
    ('securityPrivKey', c_char * USM_PRIV_KU_LEN),
    ('securityPrivKeyLen', c_size_t),
    ('securityPrivLocalKey', c_char_p),
    ('securityPrivLocalKeyLen', c_size_t),
    ('securityModel', c_int),
    ('securityLevel', c_int),
    ('paramName', c_char_p),
    ('securityInfo', c_void_p),
    ('transport_configuration', POINTER(netsnmp_container_s)),
    ('myvoid', c_void_p),
]

dataFreeHook = CFUNCTYPE(c_void_p)


class counter64(Structure):
    _fields_ = [
        ('high', c_ulong),
        ('low', c_ulong),
    ]


class netsnmp_vardata(Union):
    _fields_ = [
        ('integer', POINTER(c_long)),
        ('uinteger', POINTER(c_ulong)),
        ('string', c_char_p),
        ('objid', POINTER(oid)),
        ('bitstring', POINTER(c_ubyte)),
        ('counter64', POINTER(counter64)),
        ('floatVal', POINTER(c_float)),
        ('doubleVal', POINTER(c_double)),
    ]


class netsnmp_variable_list(Structure):
    pass


netsnmp_variable_list._fields_ = [
    ('next_variable', POINTER(netsnmp_variable_list)),
    ('name', POINTER(oid)),
    ('name_length', c_size_t),
    ('type', c_char),
    ('val', netsnmp_vardata),
    ('val_len', c_size_t),
    ('name_loc', oid * MAX_OID_LEN),
    ('buf', c_char * 40),
    ('data', c_void_p),
    ('dataFreeHook', dataFreeHook),
    ('index', c_int),
]

netsnmp_pdu._fields_ = [
    ('version', c_long),
    ('command', c_int),
    ('reqid', c_long),
    ('msgid', c_long),
    ('transid', c_long),
    ('sessid', c_long),
    ('errstat', c_long),
    ('errindex', c_long),
    ('time', c_ulong),
    ('flags', c_ulong),
    ('securityModel', c_int),
    ('securityLevel', c_int),
    ('msgParseModel', c_int),
    ('transport_data', c_void_p),
    ('transport_data_length', c_int),
    ('tDomain', POINTER(oid)),
    ('tDomainLen', c_size_t),
    ('variables', POINTER(netsnmp_variable_list)),
    ('community', c_char_p),
    ('community_len', c_size_t),
    ('enterprise', POINTER(oid)),
    ('enterprise_length', c_size_t),
    ('trap_type', c_long),
    ('specific_type', c_long),
    ('agent_addr', c_char * 4),
    ('contextEngineID', c_char_p),
    ('contextEngineIDLen', c_size_t),
    ('contextName', c_char_p),
    ('contextNameLen', c_size_t),
    ('securityEngineID', c_char_p),
    ('securityEngineIDLen', c_size_t),
    ('securityName', c_char_p),
    ('securityNameLen', c_size_t),
    ('priority', c_int),
    ('range_subid', c_int),
    ('securityStateRef', c_void_p),
]

netsnmp_pdu_p = POINTER(netsnmp_pdu)


class netsnmp_tree(Structure):
    pass


class netsnmp_enum_list(Structure):
    pass

netsnmp_enum_list._fields_ = [
    ('next', POINTER (netsnmp_enum_list)),
    ('value', c_int),
    ('label', c_char_p)
]

class netsnmp_range_list(Structure):
    pass


class netsnmp_index_list(Structure):
    pass


netsnmp_tree._fields_ = [
    ('child_list', POINTER(netsnmp_tree)),
    ('next_peer', POINTER(netsnmp_tree)),
    ('next', POINTER(netsnmp_tree)),
    ('parent', POINTER(netsnmp_tree)),
    ('label', c_char_p),
    ('subid', c_ulong),
    ('modid', c_int),
    ('number_modules', c_int),
    ('module_list', c_int_p),
    ('tc_index', c_int),
    ('type', c_int),
    ('access', c_int),
    ('status', c_int),
    ('enums', POINTER(netsnmp_enum_list)),
    ('ranges', POINTER(netsnmp_range_list)),
    ('indexes', POINTER(netsnmp_index_list)),
    ('augments', c_char_p),
    ('varbinds', POINTER(netsnmp_variable_list)),
    ('hint', c_char_p),
    ('units', c_char_p),
    ('printomat', CFUNCTYPE(c_char_p, c_size_t, c_size_t, c_int,
                            POINTER(netsnmp_variable_list), c_void_p, c_char_p,
                            c_char_p)),
    ('printer', c_void_p),
    ('description', c_char_p),
    ('reference', c_char_p),
    ('reported', c_int),
    ('defaultValue', c_char_p),
]


class netsnmp_log_message(Structure):
    pass


netsnmp_log_message_p = POINTER(netsnmp_log_message)

log_callback = CFUNCTYPE(c_int, c_int, netsnmp_log_message_p, c_void_p)

netsnmp_log_message._fields_ = [
    ('priority', c_int),
    ('msg', c_char_p),
]


class netsnmp_transport(Structure):
    pass


netsnmp_transport._fields_ = [
    ('domain', POINTER(oid)),
    ('domain_length', c_int),
    ('local', c_char_p),
    ('local_length', c_int),
    ('remote', c_char_p),
    ('remote_length', c_int),
    ('sock', c_int),
    ('flags', c_uint),
    ('data', c_void_p),
    ('data_length', c_int),
    ('msgMaxSize', c_size_t),
    ('f_recv', c_void_p),
    ('f_send', c_void_p),
    ('f_close', c_void_p),
    ('f_accept', c_void_p),
    ('f_fmtaddr', c_void_p),
]
lib.netsnmp_tdomain_transport.restype = POINTER(netsnmp_transport)

# int snmp_input(int, netsnmp_session *, int, netsnmp_pdu *, void *);
snmp_input_t = CFUNCTYPE(c_int, c_int,
                         POINTER(netsnmp_session), c_int, netsnmp_pdu_p,
                         c_void_p)


class UnknownType(Exception):
    pass


class timeval(Structure):
    _fields_ = [
        ('tv_sec', c_long),
        ('tv_usec', c_long),
    ]


#----------------------------------------------------------------------------
# lib function defs
# 
# format is:
#	'func_name' : (
#		c_char_p, # restype
#		(c_int, c_int), # argtypes
#		None, #errcheck
#	),

LIB_FUNCTIONS = {
    # main
    'snmp_pdu_create': (netsnmp_pdu_p, (), None),
    'snmp_open': (POINTER(netsnmp_session), (), None),
    'snmp_sess_open': (POINTER(netsnmp_session), (), None),
    'snmp_api_errstring': (c_char_p, (c_int, ), None),
    'snmp_errstring': (c_char_p, (c_int, ), None),
    'snmp_error':
    (None, (POINTER(netsnmp_session), c_int_p, c_int_p, c_char_p), None),
    'snmp_sess_error':
    (None, (POINTER(netsnmp_session), c_int_p, c_int_p, c_char_p), None),
    # storage space
    'netsnmp_ds_set_boolean': (c_int, (c_int, c_int, c_int), None),
    'netsnmp_ds_get_boolean': (c_int, (c_int, c_int), None),
    'netsnmp_ds_toggle_boolean': (c_int, (c_int, c_int), None),
    'netsnmp_ds_set_int': (c_int, (c_int, c_int, c_int), None),
    'netsnmp_ds_get_int': (c_int, (c_int, c_int), None),
    'netsnmp_ds_set_string': (c_int, (c_int, c_int, c_char_p), None),
    'netsnmp_ds_get_string': (c_char_p, (c_int, c_int), None),
    'netsnmp_ds_set_void': (c_int, (c_int, c_int, c_void_p), None),
    'netsnmp_ds_get_void': (c_void_p, (c_int, c_int), None),
    # mib parsing
    'init_mib': (None, (), None),
    'shutdown_mib': (None, (), None),
    'netsnmp_set_mib_directory': (None, (c_char_p), None),
    'netsnmp_get_mib_directory': (c_char_p, (), None),
    'snprint_description':
    (c_int, (c_char_p, size_t, POINTER(oid), size_t, c_int), None),
    'snprint_value': (c_int, (c_char_p, size_t, POINTER(oid), size_t,
                              POINTER(netsnmp_variable_list)), None),
    'read_mib': (POINTER(netsnmp_tree), (c_char_p), None),
    'get_tree': (POINTER(netsnmp_tree), (), None),
    'get_tree_head': (POINTER(netsnmp_tree), (), None),
}
UNSUPPORTED_FUNCTIONS = []


def setup_lib_functions():
    for k, v in LIB_FUNCTIONS.iteritems():
        if v is None:
            continue
        try:
            f = getattr(lib, k)
            f.restype = v[0]
            if isinstance(v[1], tuple) and v[1]:
                f.argtypes = v[1]
            if v[2]:
                f.errcheck = v[2]
        except:
            UNSUPPORTED_FUNCTIONS.append(k)


setup_lib_functions()


def __shutdown_mib():
    pass


def __netsnmp_set_mib_directory(path):
    if path[0] in ('-', '+'):
        b = path[0] == '-'
        path = path[1:]
        curr_path = lib.netsnmp_ds_get_string(
            NETSNMP_DS_LIBRARY_ID,
            NETSNMP_DS_LIB_MIBDIRS, )
        if b:
            a = path
            b = curr_path
        else:
            a = curr_path
            b = path

        path = os.pathsep.join(filter(None, (a, b)))
    lib.netsnmp_ds_set_string(NETSNMP_DS_LIBRARY_ID, NETSNMP_DS_LIB_MIBDIRS,
                              path)


def __netsnmp_get_mib_directory():
    lib.netsnmp_ds_get_string(
        NETSNMP_DS_LIBRARY_ID,
        NETSNMP_DS_LIB_MIBDIRS, )


def __snprint_description(buff, l, oid, s, i):
    pass


FUNCTION_REPLACEMENTS = {
    'shutdown_mib': __shutdown_mib,
    'netsnmp_set_mib_directory': __netsnmp_set_mib_directory,
    'netsnmp_get_mib_directory': __netsnmp_get_mib_directory,
    'snprint_description': __snprint_description,
}
for k in UNSUPPORTED_FUNCTIONS:
    setattr(lib, k, FUNCTION_REPLACEMENTS.get(k, None))

#----------------------------------------------------------------------------
# decoding stuff


def to_oid(l):
    if l:
        r = (oid * len(l))()
        for i, v in enumerate(l):
            r[i] = v
    else:
        r = l
    return r


def str_to_oid(s):
    if isinstance(s, basestring):
        if s[:2] == '.1':
            #s = 'iso' + s[2:]
            r = map(int, filter(None, s.split('.')))
        else:
            an_oid = (oid * MAX_OID_LEN)()
            length = c_size_t(len(an_oid))
            r = lib.get_node(s, an_oid, byref(length))
            if r:
                r = an_oid[:length.value]
            else:
                r = None
    else:
        r = s
    return r


def strs_to_oids(oids):
    return [str_to_oid(oid) for oid in oids]


def oid_to_str(o):
    if o and not isinstance(o, basestring):
        cbuff = create_string_buffer('', SPRINT_MAX_LEN)
        length = c_size_t(len(cbuff))
        an_oid = (oid * len(o))()
        for i, v in enumerate(o):
            an_oid[i] = v
        r = lib.snprint_objid(cbuff, byref(length), an_oid, len(an_oid))
        if r:
            r = cbuff[:r]
        else:
            r = ''
    else:
        r = o
    return r


def oids_to_strs(oids):
    return [oid_to_str(o) for o in oids]


def oid_to_dot(o):
    return '.' + '.'.join(map(str, str_to_oid(o)))


def oids_to_dots(oids):
    return [oid_to_dot(o) for o in oids]


DEFAULT_OID_INFO_FIELDS = [
    'label',
    'type',
    'access',
    'status',
    'hint',
    'units',
    'description',
    'reference',
    'defaultValue',

    #'subid',
    #'modid',
    #'module_list',
    #'number_modules',
    #'tc_index',
    #'augments',
    #'reference',
    #'reported',
    #'varbinds',
]


def get_oid_info(o_str,
                 info=DEFAULT_OID_INFO_FIELDS,
                 convert_status=True,
                 convert_access=True,
                 convert_type=True):
    o = to_oid(str_to_oid(o_str))
    r = None
    if o:
        tp = lib.get_tree_head()

        tp = lib.get_tree(byref(o), c_size_t(len(o)), tp)

        if tp:
            r = {}
            for i in info:
                v = getattr(tp.contents, i)

                if convert_status and i is 'status':
                    v = status_type_to_str.get(v, v)
                elif convert_type and i is 'type':
                    asn_type = oid_type_to_asn_type.get(v, None)
                    r['type_id'] = v
                    r['type_name'] = asn_type_to_name.get(asn_type,
                                                          'octet_string')
                    r['type_py'] = asn_type_to_py.get(asn_type, repr(str))
                    v = asn_type_to_str.get(asn_type, v)
                elif convert_access and i is 'access':
                    v = access_type_to_str.get(v, v)

                r[i] = v
        else:
            r = None
    return r


def get_snmp_error(session=None):
    snmp_errno = None
    if session:
        if hasattr(session, 'contents'):
            if bool(session):
                snmp_errno = session.contents.s_snmp_errno
        else:
            snmp_errno = session.s_snmp_errno
        if session is not None:
            r = lib.snmp_api_errstring(snmp_errno)
    if snmp_errno is None:
        snmp_errno = c_int.in_dll(lib, 'snmp_errno').value
        r = lib.snmp_errstring(snmp_errno)
    return r


# NOT WORKING
#def dump_tree (o=None):
##('child_list', POINTER (netsnmp_tree)),
##('next_peer', POINTER (netsnmp_tree)),
##('next', POINTER (netsnmp_tree)),
##('parent', POINTER (netsnmp_tree)),
#tp = lib.get_tree_head ()
#if o is not None:
#o = to_oid (str_to_oid (o))
#tp = lib.get_tree (byref (o), c_size_t (len (o)), tp)

#data = {}

#def scan_node (node):
#n = node.contents
#c = 0
#while n:
#print '*n', n.label
#np = n
#while np:
#print '  +np', np.label
#child_list = np.child_list
#cl = child_list and child_list.contents
#if not cl:
#data[np.label] = asn_type_to_str.get (oid_type_to_asn_type.get (np.type, None), np.type)
#elif cl.child_list and cl.child_list.contents:
#print '    -cl', cl.label
##while cl and cl.child_list:
#scan_node (cl.child_list)
##cl = cl.child_list.contents.next_peer and cl.child_list.contents.next_peer.contents

#np = np.next_peer and np.next_peer.contents
#c+=1
#n = c<4 and n.next and n.next.contents

#if tp:
#scan_node (tp)
#return data

#-------------------------------------------------------------------------


def mkoid(n):
    if n:
        oids = (oid * len(n))()
        for i, v in enumerate(n):
            oids[i] = v
    else:
        oids = None
    return oids


def decode_oid(pdu):
    return tuple(
        [pdu.val.objid[i] for i in range(pdu.val_len / sizeof(u_long))])


def decode_ip(pdu):
    return '.'.join(map(str, pdu.val.bitstring[:4]))


def decode_big_int(pdu):
    int64 = pdu.val.counter64.contents
    return (int64.high << 32L) + int64.low


def decode_string(pdu):
    if pdu.val_len:
        return string_at(pdu.val.bitstring, pdu.val_len)
    return ''


access_type_to_str = {
    MIB_ACCESS_READONLY: 'ro',
    MIB_ACCESS_READWRITE: 'rw',
    MIB_ACCESS_WRITEONLY: 'wo',
    MIB_ACCESS_NOACCESS: 'na',
    MIB_ACCESS_NOTIFY: 'notify',
    MIB_ACCESS_CREATE: 'create',
}

status_type_to_str = {
    MIB_STATUS_MANDATORY: 'mandatory',
    MIB_STATUS_OPTIONAL: 'optional',
    MIB_STATUS_OBSOLETE: 'obsolete',
    MIB_STATUS_DEPRECATED: 'deprecated',
    MIB_STATUS_CURRENT: 'current',
}

oid_type_to_asn_type = {
    TYPE_OCTETSTR: chr(ASN_OCTET_STR),
    TYPE_INTEGER: chr(ASN_INTEGER),
    TYPE_NULL: chr(ASN_NULL),
    TYPE_OBJID: chr(ASN_OBJECT_ID),
    TYPE_BITSTRING: chr(ASN_BIT_STR),
    TYPE_IPADDR: chr(ASN_IPADDRESS),
    TYPE_COUNTER: chr(ASN_COUNTER),
    TYPE_GAUGE: chr(ASN_GAUGE),
    TYPE_TIMETICKS: chr(ASN_TIMETICKS),
    TYPE_COUNTER64: chr(ASN_COUNTER64),
    TYPE_OPAQUE: chr(ASN_OPAQUE),
}
asn_type_to_oid_type = dict([(s, i) for i, s in oid_type_to_asn_type.items()])

asn_type_to_name = {
    chr(ASN_OCTET_STR): 'octet_string',
    chr(ASN_BOOLEAN): 'boolean',
    chr(ASN_INTEGER): 'integer',
    chr(ASN_NULL): 'null',
    chr(ASN_OBJECT_ID): 'object_identifier',
    chr(ASN_BIT_STR): 'bits',
    chr(ASN_IPADDRESS): 'ipaddress',
    chr(ASN_COUNTER): 'counter',
    chr(ASN_GAUGE): 'gauge',
    chr(ASN_TIMETICKS): 'timeticks',
    chr(ASN_COUNTER64): 'counter64',
    chr(ASN_FLOAT): 'float',
    chr(ASN_DOUBLE): 'double',
    chr(ASN_OPAQUE): 'opaque',
}
name_to_asn_type = dict([(s, i) for i, s in asn_type_to_name.items()])

asn_type_to_py = {
    chr(ASN_OCTET_STR): repr(str),
    chr(ASN_BOOLEAN): repr(bool),
    chr(ASN_INTEGER): repr(int),
    chr(ASN_NULL): repr(str),
    chr(ASN_OBJECT_ID): repr(str),
    chr(ASN_BIT_STR): repr(str),
    chr(ASN_IPADDRESS): repr(str),
    chr(ASN_COUNTER): repr(int),
    chr(ASN_GAUGE): repr(int),
    chr(ASN_TIMETICKS): repr(int),
    chr(ASN_COUNTER64): repr(long),
    chr(ASN_FLOAT): repr(float),
    chr(ASN_DOUBLE): repr(float),
    chr(ASN_OPAQUE): repr(str),
}
py_to_asn_type = dict([(s, i) for i, s in asn_type_to_name.items()])

asn_type_to_str = {
    chr(ASN_OCTET_STR): 's',
    chr(ASN_BOOLEAN): 'i',
    chr(ASN_INTEGER): 'i',
    chr(ASN_NULL): 'n',
    chr(ASN_OBJECT_ID): 'o',
    chr(ASN_BIT_STR): 'b',
    chr(ASN_IPADDRESS): 'a',
    chr(ASN_COUNTER): '=',
    chr(ASN_GAUGE): '=',
    chr(ASN_TIMETICKS): 't',
    chr(ASN_COUNTER64): '=',
    chr(ASN_FLOAT): '=',
    chr(ASN_DOUBLE): '=',
    chr(ASN_OPAQUE): '=',
}
str_to_asn_type = dict([(s, i) for i, s in asn_type_to_str.items()])

decoder = {
    chr(ASN_OCTET_STR): decode_string,
    chr(ASN_BOOLEAN): lambda pdu: bool(pdu.val.integer.contents.value),
    chr(ASN_INTEGER): lambda pdu: pdu.val.integer.contents.value,
    chr(ASN_NULL): lambda pdu: None,
    chr(ASN_OBJECT_ID): decode_oid,
    chr(ASN_BIT_STR): decode_string,
    chr(ASN_IPADDRESS): decode_ip,
    chr(ASN_COUNTER): lambda pdu: pdu.val.uinteger.contents.value,
    chr(ASN_GAUGE): lambda pdu: pdu.val.integer.contents.value,
    chr(ASN_TIMETICKS): lambda pdu: pdu.val.uinteger.contents.value,
    chr(ASN_COUNTER64): decode_big_int,
    chr(ASN_APP_FLOAT): lambda pdu: pdu.val.float.contents.value,
    chr(ASN_APP_DOUBLE): lambda pdu: pdu.val.double.contents.value,
}

#DEFAULT_DECODERS = [
#chr (ASN_BOOLEAN),
#chr (ASN_INTEGER),
#chr (ASN_COUNTER),
#chr (ASN_GAUGE),
#chr (ASN_TIMETICKS),
#chr (ASN_COUNTER64),
#chr (ASN_APP_FLOAT),
#chr (ASN_APP_DOUBLE),
#]
#def default_decoder (var):
##for n in DEFAULT_DECODERS:
##print 'test', ord (n)
##try:
##return decoder[n] (pdu)
##except:
##pass
##return None
#oid = [var.name[i] for i in range (var.name_length)]
#print 'default_decoder', oid
#oid = mkoid (oid)
#buff = create_string_buffer ('', 1024)
#i = lib.snprint_value (buff, len (buff), oid, len (oid), var)
#print '  result', i
#return buff.value


def decode_type(var):
    oid = [var.name[i] for i in range(var.name_length)]
    decode = decoder.get(var.type, None)
    #print ' decoder', ord(var.type), decode, map (ord, decoder.keys ()), oid
    if not decode:
        # raise UnknownType(oid, ord(var.type))
        return (oid, None)
    return oid, decode(var)


def get_result(pdu):
    result = []
    var = pdu.variables

    while var:
        var = var.contents
        #print '======', decode_type (var)
        oid, val = decode_type(var)
        result.append((tuple(oid), val))
        var = var.next_variable

    return result

def get_enum_dict(info):
    result = {}
    enums = info["enums"]

    while enums:
        enum = enums.contents
        result[enum.value] = enum.label
        enums = enum.next

    return result

def setup_library(snmp_dir=None):
    local_dir = snmp_dir or os.path.abspath(os.path.dirname(__file__))
    mibs_path = os.path.join(local_dir, 'mibs')
    persistent_path = os.path.join(local_dir, 'persist')
    config_path = os.path.join(local_dir, 'etc')

    idx = os.path.join(mibs_path, '.index')
    if os.path.exists(idx):
        try:
            os.remove(idx)
        except:
            pass

    funcs = {
        'int': lib.netsnmp_ds_set_int,
        'bool': lib.netsnmp_ds_set_boolean,
        'str': lib.netsnmp_ds_set_string,
    }
    config = {
        NETSNMP_DS_LIB_MIBDIRS: ('str', mibs_path),
        NETSNMP_DS_LIB_PERSISTENT_DIR: ('str', persistent_path),
        NETSNMP_DS_LIB_CONFIGURATION_DIR: ('str', persistent_path),
        NETSNMP_DS_LIB_DONT_PERSIST_STATE: ('bool', 1),
        NETSNMP_DS_LIB_DEFAULT_PORT: ('int', 161),
    }
    for k, v in config.iteritems():
        t, d = v
        funcs[t](NETSNMP_DS_LIBRARY_ID, k, d)

def set_mib_dir(path):
    lib.netsnmp_set_mib_directory(path)

def add_mib_dir(path, append=True):
    if append:
        path = '+' + path
    else:
        path = '-' + path

    set_mib_dir(path)

def remove_mib_dir(path):
    curr_paths = get_mib_dir().split(os.path.pathsep)
    if path in curr_paths:
        curr_paths.remove(path)
    set_mib_dir(os.path.pathsep.join(curr_paths))

def get_mib_dir():
    return lib.netsnmp_get_mib_directory()

def read_mib(name):
    tp = lib.read_mib(name)
    return bool(tp and tp.contents)

def read_module(name):
    tp = lib.netsnmp_read_module(name)
    return tp != 0

def refresh_mibs():
    lib.shutdown_mib()
    lib.init_mib()

def init_mib():
    lib.init_mib()

#----------------------------------------------------------------------------
# exceptions


class SnmpError(Exception):
    def __init__(self, src, msg, log=False):
        self.src = src
        self.msg = msg
        txt = '%s: %s' % (src, msg)

        if log:
            lib.snmp_perror(src)

        Exception.__init__(self, txt)


class SnmpTimeout(SnmpError):
    pass


class SnmpPacketError(SnmpError):
    pass


#----------------------------------------------------------------------------
# session stuff


class SNMPManager(Singleton):
    SNMP_VERSIONS = {
        '1': SNMP_VERSION_1,
        '2': SNMP_VERSION_2c,
        '2c': SNMP_VERSION_2c,
        '3': SNMP_VERSION_3,
    }

    SNMP_SELECT_SEC = 1
    SNMP_SELECT_USEC = 0
    SNMP_SELECT_BLOCK = 1

    def init_class(self,
                   name='SNMPManager',
                   log=None,
                   max_fd=1024,
                   threaded_processor=True,
                   process_sessions_sleep=0.01,
                   local_dir=None):
        self.name = name
        self.local_dir = local_dir
        self.__log = log
        self.mutex = threading.RLock()

        self.sessions = {}
        self._signal_handlers = {}

        self._quit = False

        self._fdset2list = self._fdset2list_unix
        self._snmp_read = self._snmp_read_unix
        self.bits_per = struct.calcsize(c_long._type_) * 8

        self._max_fd = max_fd
        self._fdset = c_long * (max_fd / self.bits_per)

        self.setup_config()

        self.init_logger()

        self.log(LOG_DEBUG, 'Initializing Net-SNMP v%s library.' % version)

        lib.init_snmp(name)

        self.threaded_processor = threaded_processor
        self.process_sessions_sleep = process_sessions_sleep

        if threaded_processor:
            self._process_sessions_thread = threading.Thread(
                name='snmp_process_sessions', target=self.process_sessions)
            self._process_sessions_thread.setDaemon(True)
            self._process_sessions_thread.start()
        else:
            self._process_sessions_thread = None

        self.refresh_mibs()

    @syncronized
    def destroy(self, shutdown=True):
        if self.destroyed:
            return

        self._quit = True
        if self._process_sessions_thread and self._process_sessions_thread.isAlive(
        ):
            self._process_sessions_thread.join(5)
            if self._process_sessions_thread.isAlive():
                self._process_sessions_thread.kill()
            self._process_sessions_thread = None

        self.destroy_sessions()

        if shutdown:
            self.log(LOG_DEBUG, 'Shutting down Net-SNMP v%s library.' %
                     version)

            s = ' ' * 255
            cbuff = create_string_buffer(s, len(s))
            lib.snmp_shutdown(cbuff)

        self.destroy_logger()

        self._signal_handlers.clear()
        self.__log = None
        self.mutex = None

        self._snmp_read = None
        self._fdset = None
        self._fdset2list = None

        Singleton.destroy(self)

    # config ----------------------------------------------------------------

    def setup_config(self):
        setup_library(snmp_dir=self.local_dir)

    # signals ----------------------------------------------------------------

    def emit(self, slot, session, reqid, *args, **kargs):
        if slot in self._signal_handlers:
            handlers = []
            handlers.extend(self._signal_handlers[slot].get(session, {})
                            .values())
            handlers.extend(self._signal_handlers[slot].get(None, {}).values())
            for cb in (h['callback'] for h in handlers):
                try:
                    cb(self, slot, session, reqid, *args, **kargs)
                except Exception, e:
                    self.log(LOG_DEBUG, 'callback exception: %s' % str(e))
                    self.log(LOG_DEBUG,
                             traceback.format_exc(sys.exc_info()[2]))

    def bind(self, slot, uid, session, callback):
        self._signal_handlers.setdefault(slot, {}).setdefault(
            session, {})[uid] = {
                'callback': callback
            }

    def unbind(self, slot, uid, session=None):
        if slot in self._signal_handlers:
            for k, v in self._signal_handlers[slot].iteritems():
                if session is None or session == k:
                    if uid in v:
                        del v[uid]

    def process_sessions(self):
        while not self._quit:
            self.mutex.acquire()
            rd, t = self._snmp_select_info(self.SNMP_SELECT_SEC,
                                           self.SNMP_SELECT_USEC,
                                           self.SNMP_SELECT_BLOCK)
            if t is None:
                if not self.threaded_processor:
                    break

            if t is not None:
                rd, wd, xd = select(rd, [], [], t)
                if rd:
                    self._snmp_read(rd)
                else:
                    lib.snmp_timeout()

            self.timeout_async_requests()
            self.mutex.release()

            if self.process_sessions_sleep:
                time.sleep(self.process_sessions_sleep)

    def timeout_async_requests(self):
        now = time.time()

        for i in self.sessions.values():
            i.timeout_async_requests(now)

    def _snmp_select_info(self, sec=1, usec=0, block=0):
        timeout = timeval()
        timeout.tv_sec = sec
        timeout.tv_usec = usec

        f_block = c_int(0)

        rd = self._fdset()
        maxfd = c_int(0)  #self._max_fd)

        cnt = lib.snmp_select_info(
            byref(maxfd), byref(rd), byref(timeout), byref(f_block))
        rd = self._fdset2list(rd, maxfd.value, cnt)

        if block and f_block:
            t = None
        elif block and not f_block:
            t = timeout.tv_sec + float(timeout.tv_usec) / 1e6
        else:
            t = 0
        return rd, t

    def _fdset2list_unix(self, rd, n, cnt):
        result = []
        #for i in range (cnt):
        for i in range(len(rd)):
            if rd[i]:
                for j in range(0, self.bits_per):
                    bit = 0x00000001 << (j % self.bits_per)
                    if rd[i] & bit:
                        result.append(i * self.bits_per + j)
        return result

    def _snmp_read_unix(self, d):
        for fd in d:
            rd = self._fdset()
            rd[fd / self.bits_per] |= 1 << (fd % self.bits_per)
            lib.snmp_read(byref(rd))

    # mibs ----------------------------------------------------------------

    def set_mib_dir(self, path):
        lib.netsnmp_set_mib_directory(path)

    def add_mib_dir(self, path, append=True):
        add_mib_dir(path, append)

    def remove_mib_dir(self, path):
        remove_mib_dir(path)

    def get_mib_dir(self):
        return get_mib_dir()

    def read_mib(self, name):
        return read_mib(name)

    def read_module(self, name):
        return read_module(name)

    def refresh_mibs(self):
        return refresh_mibs()

    def init_mib(self):
        return init_mib()

    # logging ----------------------------------------------------------------

    PRIORITY_MAP = {
        LOG_EMERG: 'emergency',
        LOG_ALERT: 'alert',
        LOG_CRIT: 'critical',
        LOG_ERR: 'error',
        LOG_WARNING: 'warning',
        LOG_NOTICE: 'notice',
        LOG_INFO: 'info',
        LOG_DEBUG: 'debug',
    }

    def log(self, priority, msg):
        priority = self.PRIORITY_MAP.get(priority, priority)
        if self.__log:
            self.__log(priority, msg)
        else:
            print priority, ':', msg

    def _netsnmp_logger(self, a, b, msg):
        msg = cast(msg, netsnmp_log_message_p)
        priority = self.PRIORITY_MAP.get(msg.contents.priority, 'warning')
        if self.__log:
            self.__log(priority, msg.contents.msg.strip())
        else:
            print priority, ':', msg.contents.msg.strip()
        return 0

    def init_logger(self):
        self._netsnmp_logger_callback = log_callback(self._netsnmp_logger)

        lib.snmp_register_callback(SNMP_CALLBACK_LIBRARY,
                                   SNMP_CALLBACK_LOGGING,
                                   self._netsnmp_logger_callback, 0)

        self._log_handler = lib.netsnmp_register_loghandler(
            NETSNMP_LOGHANDLER_CALLBACK, LOG_DEBUG)

    def destroy_logger(self):
        # commented becouse errors are generated on win32
        #lib.netsnmp_remove_loghandler (self._log_handler)
        self._netsnmp_logger_callback = None
        self._log_handler = None

    # sessions ----------------------------------------------------------------

    def add_session(self, name, version='1', **kargs):
        if not version in self.SNMP_VERSIONS:
            raise RuntimeError('Unknown SNMP version "%s"' % version)

        if 'community' in kargs and 'community_len' not in kargs:
            kargs['community_len'] = len(kargs['community'])
        if 'timeout' in kargs:
            if kargs['timeout'] is None:
                kargs['timeout'] = SNMP_DEFAULT_TIMEOUT
            else:
                kargs['timeout'] = int(kargs['timeout'] * 1e6)
        if 'retries' in kargs:
            if kargs['retries'] is None:
                kargs['retries'] = SNMP_DEFAULT_RETRIES

        s = SNMPSession(
            manager=self,
            name=name,
            version=self.SNMP_VERSIONS[version],
            **kargs)

        try:
            s.open()
            self.sessions[name] = s
        except SnmpError:
            s.destroy()
            s = None
            raise
        return s

    def add_trapd_session(self, name, peername, fileno=-1):
        s = SNMPSession(manager=self, name=name)
        try:
            s.trapd(peername, fileno)
            self.sessions[name] = s
        except SnmpError:
            s.destroy()
            s = None
            raise
        return s

    def remove_session(self, name):
        if name in self.sessions:
            s = self.sessions[name]
            del self.sessions[name]
            s.destroy()

    def reopen_sessions(self):
        for i in self.sessions.values():
            i.close()
            i.open()
        self.sessions.clear()

    def destroy_sessions(self):
        for i in self.sessions.values():
            i.destroy()
        self.sessions.clear()

    def find_session(self, sessid):
        r = None
        for v in self.sessions.values():
            if sessid == v.sess.contents.sessid:
                r = v
                break
        return r

    def __getitem__(self, name):
        return self.sessions[name]


class SNMPSession(object):
    ASYNC_REQUEST_TIMEOUT = 60

    def __init__(self, manager, name, results_as_list=False, **kargs):
        self.manager = manager
        self.name = name
        self.results_as_list = results_as_list
        self.kw = kargs
        self.mutex = threading.RLock()
        self.async_requests = {}
        self.async_wait_list = {}
        self.async_request_timeout = self.ASYNC_REQUEST_TIMEOUT

    def destroy(self):
        self.manager.mutex.acquire()
        self.close()
        self.manager.mutex.release()
        self.manager = None
        self.name = None
        self.kw = None
        self.sess = None
        self.mutex = None
        self._netsnmp_callback = None

    # requests --------------------------------------------------------------

    def open(self):
        sess = netsnmp_session()
        lib.snmp_sess_init(byref(sess))

        for attr, value in self.kw.items():
            setattr(sess, attr, value)

        sess.callback = self._netsnmp_callback  #self._get_callback ()
        sess.callback_magic = id(self)
        #lib.snmp_set_do_debugging (1)
        self.sess = lib.snmp_open(byref(sess))
        #lib.snmp_set_do_debugging (0)

        if not self.sess:
            raise SnmpError('snmp_open', get_snmp_error(sess))

    def close(self):
        if self.sess:
            lib.snmp_close(self.sess)
            self.sess = None

        self.async_requests.clear()

        for i in self.async_wait_list.values():
            if isinstance(i, threading._Event):
                i.set()
        self.async_wait_list.clear()

    # traps --------------------------------------------------------------

    def trapd(self, peername, fileno=-1):
        lib.netsnmp_udp_ctor()

        transport = lib.netsnmp_tdomain_transport(peername, 1, 'udp')
        if not transport:
            raise SnmpError('Unable to create transport', peername)

        if fileno >= 0:
            os.dup2(fileno, transport.contents.sock)

        sess = netsnmp_session()
        self.sess = lib.snmp_sess_init(byref(sess))
        if not self.sess:
            raise SnmpError(name, 'snmp_sess_init')

        sess.peername = SNMP_DEFAULT_PEERNAME
        sess.version = SNMP_DEFAULT_VERSION
        sess.community_len = SNMP_DEFAULT_COMMUNITY_LEN
        sess.retries = SNMP_DEFAULT_RETRIES
        sess.timeout = SNMP_DEFAULT_TIMEOUT
        sess.callback = self._netsnmp_callback
        sess.callback_magic = id(self)
        # sess.authenticator = None
        sess.isAuthoritative = SNMP_SESS_UNKNOWNAUTH
        rc = lib.snmp_add(self.sess, transport, None, None)
        if not rc:
            raise SnmpError(name, 'snmp_add')

    # utils ------------------------------------------------------

    def _create_request(self, packet_type):
        return lib.snmp_pdu_create(packet_type)

    def pdu_parse(self, pdu, buffer):
        cbuff = create_string_buffer(buffer, len(buffer))
        length = c_size_t(len(buffer))
        after_header = c_char_p()
        err = lib.snmpv3_parse(
            byref(pdu), cbuff, byref(length), byref(after_header), self.sess)
        if err:
            raise SnmpError('pdu_parse', lib.snmp_errstring(err))

    # info ------------------------------------------------------

    def get_description(self, oid, width=80, buffer_size=10240):
        oid = str_to_oid(oid)
        if oid is not None:
            oid = mkoid(oid)
            buff = create_string_buffer('', buffer_size)
            lib.snprint_description(buff, len(buff), oid, len(oid), width)
            r = buff.value
        else:
            r = ''
        return r

    # sync operations ------------------------------------------------------

    #def _handle_sync_request (self, name, req, exc_on_error):
    #response = netsnmp_pdu_p ()
    #status = lib.snmp_synch_response (self.sess, req, byref (response))
    #result = None
    #r = bool (response) and response.contents
    #try:
    #if status == STAT_SUCCESS:
    #if r and r.errstat == SNMP_ERR_NOERROR:
    #result = dict (get_result (r))
    #else:
    #if exc_on_error:
    #raise SnmpPacketError (name)
    #elif r:
    #result = lib.snmp_errstring (r.errstat)
    #elif status == STAT_TIMEOUT:
    #result = 'timeout'
    #if exc_on_error:
    #raise SnmpTimeout (name, result)
    #else: # STAT_ERROR
    #if r:
    #result = lib.snmp_errstring (r.errstat)
    #else:
    #result = get_snmp_error (self.sess) or 'Unknown Error'
    #if exc_on_error:
    #raise SnmpError (name, result)
    #finally:
    #if response:
    #lib.snmp_free_pdu (response)

    #return result

    def sync_get(self, oids, exc_on_error=False):
        return self.wait_async_request(
            self.async_get(oids, True, exc_on_error))

        #oids = strs_to_oids (oids)
        #req = self._create_request (SNMP_MSG_GET)

        #for oid in oids:

    #if oid is not None:
    #oid = mkoid (oid)
    #lib.snmp_add_null_var (req, oid, len (oid))
    #return self._handle_sync_request ('sync_get', req, exc_on_error)

    def sync_getbulk(self,
                     nonrepeaters,
                     maxrepetitions,
                     oids,
                     exc_on_error=False):
        return self.wait_async_request(
            self.async_getbulk(nonrepeaters, maxrepetitions, oids, True,
                               exc_on_error))

        #oids = strs_to_oids (oids)
        #req = self._create_request (SNMP_MSG_GETBULK)
        #req.contents.errstat = nonrepeaters
        #req.contents.errindex = maxrepetitions
        #for oid in oids:
    #if oid is not None:
    #oid = mkoid (oid)
    #lib.snmp_add_null_var (req, oid, len (oid))

    #return self._handle_sync_request ('sync_getbulk', req, exc_on_error)

    def sync_walk(self, root, exc_on_error=False):
        return self.wait_async_request(
            self.async_walk(root, True, exc_on_error))

        #req = self._create_request (SNMP_MSG_GETNEXT)
        #oid = mkoid (str_to_oid (root))
        #if oid is not None:

    #lib.snmp_add_null_var (req, oid, len (oid))

    #return self._handle_sync_request ('sync_getnext', req, exc_on_error)

    # async operations ------------------------------------------------------

    # had to implement the hack below becouse the _get_callback () method doesn't
    # work under OS X (why?)

    def _netsnmp_callback(operation, sp, reqid, pdu, magic):
        s = SNMPManager().find_session(sp.contents.sessid)
        if s:
            r = s.dispatch_callback(operation, sp, reqid, pdu, magic)
        else:
            r = 1
        return r
    _netsnmp_callback = netsnmp_callback(_netsnmp_callback)

    @syncronized
    def dispatch_callback(self, operation, sp, reqid, pdu, magic):
        try:
            if operation == NETSNMP_CALLBACK_OP_RECEIVED_MESSAGE:
                self.callback(pdu)
            elif operation == NETSNMP_CALLBACK_OP_TIMED_OUT:
                self.timeout(reqid)
            else:
                self._process_waiting_async_request(
                    reqid, "Unknown operation (%d)" % operation)
                self.manager.log(LOG_ERR, "Unknown operation (%d)" % operation)
        except Exception, ex:
            self._process_waiting_async_request(
                reqid, "Exception in dispatch_callback (%s)" % ex)
            self.manager.log(LOG_ERR, "Exception in dispatch_callback (%s)" %
                             ex)

        return 1

    #def _get_callback (self):
    #def _callback (operation, sp, reqid, pdu, magic):
    #try:
    #if operation == NETSNMP_CALLBACK_OP_RECEIVED_MESSAGE:
    #self.callback (pdu)
    #elif operation == NETSNMP_CALLBACK_OP_TIMED_OUT:
    #self.timeout (reqid)
    #else:
    #self._process_waiting_async_request (reqid, "Unknown operation (%d)" % operation)
    #self.manager.log (LOG_ERR, "Unknown operation (%d)" % operation)
    #except Exception, ex:
    ##self._process_waiting_async_request (reqid, "Exception in _callback (%s)" % ex)
    #self.manager.log (LOG_ERR, "Exception in _callback (%s)" % ex, ex)

    #return 1
    #_callback = netsnmp_callback (_callback)
    #return _callback

    COMMAND_TO_SLOT = {
        SNMP_MSG_GET: 'get',
        SNMP_MSG_GETNEXT: 'getnext',
        SNMP_MSG_GETBULK: 'getbulk',
        SNMP_MSG_SET: 'set',
        SNMP_MSG_INFORM: 'inform',
        SNMP_MSG_TRAP: 'trap',
        SNMP_MSG_TRAP2: 'trap2',
        SNMP_MSG_REPORT: 'report',
        SNMP_MSG_RESPONSE: 'response',
    }
    SLOT_TO_COMMAND = dict([(v, k) for k, v in COMMAND_TO_SLOT.iteritems()])

    def callback(self, pdu):
        reqid = pdu.contents.reqid
        if self.results_as_list:
            result = get_result(pdu.contents)
        else:
            result = dict(get_result(pdu.contents))

        if reqid in self.async_requests:
            rtype, timeout = self.async_requests.pop(reqid, (None, None))
            slot = self.COMMAND_TO_SLOT.get(rtype, None)
            b = self._process_waiting_async_request(reqid, result)
        else:
            slot = self.COMMAND_TO_SLOT.get(pdu.contents.command, 'result')
            b = False

        if b:
            if slot:
                self.manager.emit(slot, self.name, reqid, result)
                self.manager.emit('response', self.name, reqid, result)

    def timeout(self, reqid):
        if self._process_waiting_async_request(reqid, 'timeout'):
            self.manager.emit('timeout', self.name, reqid)

    @syncronized
    def _async_send_request(self, name, req, wait, exc_on_error):
        if not lib.snmp_send(self.sess, req):
            lib.snmp_free_pdu(req)
            result = get_snmp_error(self.sess)
            if exc_on_error:
                raise SnmpError(name, err)
        else:
            timeout = time.time() + self.async_request_timeout
            reqid = req.contents.reqid
            self.async_requests[reqid] = (name, timeout)
            if wait:
                e = threading.Event()
                e.timeout = timeout
                self.async_wait_list[reqid] = e
                result = (reqid, e)
            else:
                result = reqid
        return result

    def _process_waiting_async_request(self, reqid, result):
        b = True
        if reqid in self.async_wait_list:
            if isinstance(self.async_wait_list[reqid], threading._Event):
                e = self.async_wait_list[reqid]
                self.async_wait_list[reqid] = result
                e.set()
                b = False
        return b

    @syncronized
    def timeout_async_requests(self, now):
        for reqid in self.async_wait_list.keys():
            if isinstance(self.async_wait_list[reqid], threading._Event):
                e = self.async_wait_list[reqid]
                if e.timeout < now:
                    self.async_wait_list[reqid] = 'timeout'
                    e.set()

        for reqid in self.async_requests.keys():
            if self.async_requests[reqid][1] < now:
                del self.async_requests[reqid]

    def wait_async_request(self, result):
        reqid, e = result
        e.wait()
        self.mutex.acquire()
        try:
            result = self.async_wait_list.pop(reqid, None)
        finally:
            self.mutex.release()
        return result

    def async_get(self, oids, wait=False, exc_on_error=False):
        oids = strs_to_oids(oids)
        req = self._create_request(SNMP_MSG_GET)
        for oid in oids:
            if oid is not None:
                oid = mkoid(oid)
                lib.snmp_add_null_var(req, oid, len(oid))

        return self._async_send_request(SNMP_MSG_GET, req, wait, exc_on_error)

    def async_getbulk(self,
                      nonrepeaters,
                      maxrepetitions,
                      oids,
                      wait=False,
                      exc_on_error=False):
        req = self._create_request(SNMP_MSG_GETBULK)
        req.contents.errstat = nonrepeaters
        req.contents.errindex = maxrepetitions

        oids = strs_to_oids(oids)
        for oid in oids:
            if oid is not None:
                oid = mkoid(oid)
                lib.snmp_add_null_var(req, oid, len(oid))

        return self._async_send_request(SNMP_MSG_GETBULK, req, wait,
                                        exc_on_error)

    def async_walk(self, root, wait=False, exc_on_error=False):
        req = self._create_request(SNMP_MSG_GETNEXT)
        oid = mkoid(str_to_oid(root))
        if oid is not None:
            lib.snmp_add_null_var(req, oid, len(oid))

        return self._async_send_request(SNMP_MSG_GETNEXT, req, wait,
                                        exc_on_error)

    # set operations ------------------------------------------------------

    def _set(self, oids_values):
        req = self._create_request(SNMP_MSG_SET)

        for oid, value in oids_values:
            oid = str_to_oid(oid)
            if oid is not None:
                oid_type = get_oid_info(oid, ['type']).get('type', '=')
                if not oid_type:
                    oid_type = '='
                oid = mkoid(oid)
                lib.snmp_add_var(req, oid, len(oid), ord(oid_type), value)
        return req

    def sync_set(self, oids_values, exc_on_error=False):
        return self.wait_async_request(
            self.async_set(oids_values, True, exc_on_error))

        #req = self._set (oids_values, exc_on_error)
        #return self._handle_sync_request ('sync_set', req, exc_on_error)

    def async_set(self, oids_values, wait=False, exc_on_error=False):
        req = self._set(oids_values)
        return self._async_send_request(SNMP_MSG_SET, req, wait, exc_on_error)


if __name__ == '__main__':
    # Some stupid tests
    for t in range(1):
        print 'init manager'
        m = SNMPManager()

        print 'setup mib dirs'
        #m.add_mib_dir (os.path.abspath (os.path.join (os.path.basename (__file__), 'mibs')))
        #m.refresh_mibs ()
        print 'read_mib', m.read_mib(
            os.path.join('test', 'data', 'LINKSYS.MIB'))
        print 'mib dirs:', m.get_mib_dir()

        PEERNAME = 'localhost'
        SNMP_VERSION = '2'
        PUBLIC = 'public'
        PRIVATE = 'private'

        print 'add session'
        try:
            s = m.add_session(
                name='get_session',
                version=SNMP_VERSION,
                peername=PEERNAME,
                community=PUBLIC)
            set_s = m.add_session(
                name='set_session',
                version=SNMP_VERSION,
                peername=PEERNAME,
                community=PRIVATE)
        except SnmpError, exc:
            print 'ABORT on open session', exc
            m.destroy()
            sys.exit()

        mibs = [
            'sysDescr.0',
            'sysObjectID.0',
            'sysLocation.0',
            '.1.3.6.1.2.1.1.1.0',
            'sysUpTimeInstance',
            'ifOutOctets.1',
            'laLoadFloat.0',
            'snmpInTotalReqVars.0',
            'hrSystemProcesses.0',
        ]

        print 'get descriptions'
        for i in mibs:
            print i, '=' * 40
            print(i, str_to_oid(i), oid_to_str(str_to_oid(i)), get_oid_info(i)
                  )  #['type'])
            print s.get_description(i)

        print '- sync get'
        for i in range(1):
            r = s.sync_get(mibs)
            print '-' * 40
            if isinstance(r, dict):
                for k, v in r.iteritems():
                    try:
                        print '%s (%s) = %s' % (oid_to_str(k), k, v)
                    except:
                        #Failure ().print_brief_traceback ()
                        pass
            else:
                print 'ERROR:', r
        #m.destroy ()
        #sys.exit ()

        print '- sync getbulk'
        for i in range(1):
            r = s.sync_getbulk(0, 10, mibs)
            print '-' * 40
            if isinstance(r, dict):
                for k, v in r.iteritems():
                    try:
                        print '%s (%s) = %s' % (oid_to_str(k), k, v)
                    except:
                        #Failure ().print_brief_traceback ()
                        pass
            else:
                print 'ERROR:', r

        print '- sync walk'
        o = 'SNMPv2-MIB'
        for i in range(3):
            r = s.sync_walk(o)
            print '-' * 40
            if isinstance(r, dict):
                for k, v in r.iteritems():
                    o = k
                    try:
                        print '%s (%s) = %s' % (oid_to_str(k), k, v)
                    except:
                        #Failure ().print_brief_traceback ()
                        pass
            else:
                print 'ERROR:', r

        # ASYNC STUFF

        def timeout_cb(manager, slot, session, reqid):
            print '!', 'async timeout "%s"' % slot, (manager, session, reqid)

        def result_cb(manager, slot, session, reqid, r):
            print '=' * 40, 'async "%s"' % slot, (manager, session, reqid)
            if r:
                for k, v in r.iteritems():
                    o = k
                    try:
                        print '%s (%s) = %s' % (oid_to_str(k), k, v)
                    except:
                        #Failure ().print_brief_traceback ()
                        pass

        m.bind('response', '1', None, result_cb)
        m.bind('timeout', '1', None, timeout_cb)

        print '- async get',
        for i in range(1):
            r = s.async_get(mibs)
            print r
        print '* wait 5s'
        time.sleep(5)
        print '- async getbulk',
        for i in range(1):
            r = s.async_getbulk(0, 10, mibs)
            print r
        print '* wait 5s'
        time.sleep(5)
        print '- async walk',
        o = 'SNMPv2-MIB'
        r = s.async_walk(o)
        print r

        print '* wait 5s'
        time.sleep(5)
        print '- sync_set', set_s.sync_get(['sysLocation.0'])
        print '+ set result = ', set_s.sync_set([('sysLocation.0', 'Plaino')])
        print '- Udine = ', set_s.sync_get(['sysLocation.0'])
        print '* wait 5s'
        time.sleep(5)
        print '- async_set', set_s.async_get(['sysLocation.0'])
        print '+ set result = ', set_s.async_set([('sysLocation.0', 'Udine')])
        time.sleep(5)
        print '- Udine = ', set_s.async_get(['sysLocation.0'])
        print '* wait 5s'
        time.sleep(5)

        print 'destroy manager'
        m.unbind('response', '1')
        m.destroy()



#!/usr/bin/env python
"""
Setup file for fusnmp distribution.
"""
import sys
from setuptools import setup

VERSION = '1.0.3'

try:
    # Use pandoc to convert .md -> .rst when uploading to pypi
    import pypandoc
    DESCRIPTION = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError, OSError):
    DESCRIPTION = open('README.md').read()

if sys.version_info[0] == 2 and sys.version_info[1] < 7:
    sys.exit('Sorry, Python 2 < 2.7 is not supported')

if sys.version_info[0] == 3:
    sys.exit('Sorry, Python 3 < 3.3 is not supported')

setup(
    name='fusnmp',
    version=VERSION,
    description="A small but complete NetSNMP ctypes wrapper.",
    long_description=DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords='snmp networking',
    author='Jon M. Dugan',
    author_email='jdugan@x1024.net',
    url='http://github.com/jdugan1024/fusnmp/',
    license='GPL',
    packages=["fusnmp"],
    classifiers=[
        'Development Status :: 7 - Inactive',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Topic :: Internet',
        'Operating System :: POSIX',
    ],
)
